@doc raw"""
Abstract supertype of all predictive controllers.

---

    (mpc::PredictiveController)(ry, d=[]; kwargs...) -> u

Functor allowing callable `PredictiveController` object as an alias for [`moveinput!`](@ref).

# Examples
```jldoctest
julia> mpc = LinMPC(LinModel(tf(5, [2, 1]), 3), Nwt=[0], Hp=1000, Hc=1);

julia> u = mpc([5]); round.(u, digits=3)
1-element Vector{Float64}:
 1.0
```

"""
abstract type PredictiveController end

const DEFAULT_HP0 = 10
const DEFAULT_HC  = 2
const DEFAULT_MWT = 1.0
const DEFAULT_NWT = 0.1
const DEFAULT_LWT = 0.0
const DEFAULT_CWT = 1e5
const DEFAULT_EWT = 0.0

"Termination status that means 'no solution available'."
const FATAL_STATUSES = [
    INFEASIBLE, DUAL_INFEASIBLE, LOCALLY_INFEASIBLE, INFEASIBLE_OR_UNBOUNDED, 
    NUMERICAL_ERROR, INVALID_MODEL, INVALID_OPTION, INTERRUPTED, 
    OTHER_ERROR
]

"Type alias for vector of linear inequality constraints."
const LinConVector = Vector{ConstraintRef{
    Model, 
    MOI.ConstraintIndex{MOI.ScalarAffineFunction{Float64}, MOI.LessThan{Float64}}, 
    ScalarShape
}}

const InfoDictType = Union{JuMP._SolutionSummary, Vector{Float64}, Float64}

"Include all the data for the constraints of [`PredictiveController`](@ref)"
struct ControllerConstraint
    ẽx̂      ::Matrix{Float64}
    fx̂      ::Vector{Float64}
    gx̂      ::Matrix{Float64}
    jx̂      ::Matrix{Float64}
    kx̂      ::Matrix{Float64}
    vx̂      ::Matrix{Float64}
    Umin    ::Vector{Float64}
    Umax    ::Vector{Float64}
    ΔŨmin   ::Vector{Float64}
    ΔŨmax   ::Vector{Float64}
    Ymin    ::Vector{Float64}
    Ymax    ::Vector{Float64}
    x̂min    ::Vector{Float64}
    x̂max    ::Vector{Float64}
    A_Umin  ::Matrix{Float64}
    A_Umax  ::Matrix{Float64}
    A_ΔŨmin ::Matrix{Float64}
    A_ΔŨmax ::Matrix{Float64}
    A_Ymin  ::Matrix{Float64}
    A_Ymax  ::Matrix{Float64}
    A_x̂min  ::Matrix{Float64}
    A_x̂max  ::Matrix{Float64}
    A       ::Matrix{Float64}
    b       ::Vector{Float64}
    i_b     ::BitVector
    C_ymin  ::Vector{Float64}
    C_ymax  ::Vector{Float64}
    c_x̂min  ::Vector{Float64}
    c_x̂max  ::Vector{Float64}
    i_g     ::BitVector
end

@doc raw"""
    setconstraint!(mpc::PredictiveController; <keyword arguments>) -> mpc

Set the constraint parameters of `mpc` predictive controller.

The predictive controllers support both soft and hard constraints, defined by:
```math 
\begin{alignat*}{3}
    \mathbf{u_{min}  - c_{u_{min}}}  ϵ ≤&&\       \mathbf{u}(k+j) &≤ \mathbf{u_{max}  + c_{u_{max}}}  ϵ &&\qquad  j = 0, 1 ,..., H_p - 1 \\
    \mathbf{Δu_{min} - c_{Δu_{min}}} ϵ ≤&&\      \mathbf{Δu}(k+j) &≤ \mathbf{Δu_{max} + c_{Δu_{max}}} ϵ &&\qquad  j = 0, 1 ,..., H_c - 1 \\
    \mathbf{y_{min}  - c_{y_{min}}}  ϵ ≤&&\       \mathbf{ŷ}(k+j) &≤ \mathbf{y_{max}  + c_{y_{max}}}  ϵ &&\qquad  j = 1, 2 ,..., H_p     \\
    \mathbf{x̂_{min}  - c_{x̂_{min}}}  ϵ ≤&&\ \mathbf{x̂}_{k-1}(k+j) &≤ \mathbf{x̂_{max}  + c_{x̂_{max}}}  ϵ &&\qquad  j = H_p
\end{alignat*}
```
and also ``ϵ ≥ 0``. The last line is the terminal constraints applied on the states at the
end of the horizon (see Extended Help). All the constraint parameters are vector. Use `±Inf`
values when there is no bound. The constraint softness parameters ``\mathbf{c}``, also
called equal concern for relaxation, are non-negative values that specify the softness of
the associated bound. Use `0.0` values for hard constraints. The output and terminal 
constraints are all soft by default. See Extended Help for time-varying constraints.

# Arguments
!!! info
    The default constraints are mentioned here for clarity but omitting a keyword argument 
    will not re-assign to its default value (defaults are set at construction only).

- `mpc::PredictiveController` : predictive controller to set constraints.
- `umin  = fill(-Inf,nu)` : manipulated input lower bounds ``\mathbf{u_{min}}``.
- `umax  = fill(+Inf,nu)` : manipulated input upper bounds ``\mathbf{u_{max}}``.
- `Δumin = fill(-Inf,nu)` : manipulated input increment lower bounds ``\mathbf{Δu_{min}}``.
- `Δumax = fill(+Inf,nu)` : manipulated input increment upper bounds ``\mathbf{Δu_{max}}``.
- `ymin  = fill(-Inf,ny)` : predicted output lower bounds ``\mathbf{y_{min}}``.
- `ymax  = fill(+Inf,ny)` : predicted output upper bounds ``\mathbf{y_{max}}``.
- `x̂min  = fill(-Inf,nx̂)` : terminal constraint lower bounds ``\mathbf{x̂_{min}}``.
- `x̂max  = fill(+Inf,nx̂)` : terminal constraint upper bounds ``\mathbf{x̂_{max}}``.
- `c_umin  = fill(0.0,nu)` : `umin` softness weights ``\mathbf{c_{u_{min}}}``.
- `c_umax  = fill(0.0,nu)` : `umax` softness weights ``\mathbf{c_{u_{max}}}``.
- `c_Δumin = fill(0.0,nu)` : `Δumin` softness weights ``\mathbf{c_{Δu_{min}}}``.
- `c_Δumax = fill(0.0,nu)` : `Δumax` softness weights ``\mathbf{c_{Δu_{max}}}``.
- `c_ymin  = fill(1.0,ny)` : `ymin` softness weights ``\mathbf{c_{y_{min}}}``.
- `c_ymax  = fill(1.0,ny)` : `ymax` softness weights ``\mathbf{c_{y_{max}}}``.
- `c_x̂min  = fill(1.0,nx̂)` : `x̂min` softness weights ``\mathbf{c_{x̂_{min}}}``.
- `c_x̂max  = fill(1.0,nx̂)` : `x̂max` softness weights ``\mathbf{c_{x̂_{max}}}``.
- all the keyword arguments above but with a capital letter, except for the terminal
  constraints, e.g. `Ymax` or `C_Δumin` : for time-varying constraints (see Extended Help).

# Examples
```jldoctest
julia> mpc = LinMPC(setop!(LinModel(tf(3, [30, 1]), 4), uop=[50], yop=[25]));

julia> mpc = setconstraint!(mpc, umin=[0], umax=[100], c_umin=[0.0], c_umax=[0.0]);

julia> mpc = setconstraint!(mpc, Δumin=[-10], Δumax=[+10], c_Δumin=[1.0], c_Δumax=[1.0])
LinMPC controller with a sample time Ts = 4.0 s, OSQP optimizer, SteadyKalmanFilter estimator and:
 10 prediction steps Hp
  2 control steps Hc
  1 manipulated inputs u (0 integrating states)
  2 states x̂
  1 measured outputs ym (1 integrating states)
  0 unmeasured outputs yu
  0 measured disturbances d
```

# Extended Help
Terminal constraints provide closed-loop stability guarantees on the nominal plant model.
They can render an unfeasible problem however. In practice, a sufficiently large prediction
horizon ``H_p`` without terminal constraints is typically enough for stability. Note that
terminal constraints are applied on the augmented state vector ``\mathbf{x̂}`` (see
[`SteadyKalmanFilter`](@ref) for details on augmentation).

For variable constraints, the bounds can be modified after calling [`moveinput!`](@ref),
that is, at runtime, but not the softness parameters ``\mathbf{c}``. It is not possible to
modify `±Inf` bounds at runtime.

!!! tip
    To keep a variable unconstrained while maintaining the ability to add a constraint later
    at runtime, set the bound to an absolute value sufficiently large when you create the
    controller (but different than `±Inf`).

It is also possible to specify time-varying constraints over ``H_p`` and ``H_c`` horizons. 
In such a case, they are defined by:
```math 
\begin{alignat*}{3}
    \mathbf{U_{min}  - C_{u_{min}}}  ϵ ≤&&\ \mathbf{U}  &≤ \mathbf{U_{max}  + C_{u_{max}}}  ϵ \\
    \mathbf{ΔU_{min} - C_{Δu_{min}}} ϵ ≤&&\ \mathbf{ΔU} &≤ \mathbf{ΔU_{max} + C_{Δu_{max}}} ϵ \\
    \mathbf{Y_{min}  - C_{y_{min}}}  ϵ ≤&&\ \mathbf{Ŷ}  &≤ \mathbf{Y_{max}  + C_{y_{max}}}  ϵ
\end{alignat*}
```
For this, use the same keyword arguments as above but with a capital letter:
- `Umin`  / `Umax`  / `C_umin`  / `C_umax`  : ``\mathbf{U}`` constraints `(nu*Hp,)`.
- `ΔUmin` / `ΔUmax` / `C_Δumin` / `C_Δumax` : ``\mathbf{ΔU}`` constraints `(nu*Hc,)`.
- `Ymin`  / `Ymax`  / `C_ymin`  / `C_ymax`  : ``\mathbf{Ŷ}`` constraints `(ny*Hp,)`.
"""
function setconstraint!(
    mpc::PredictiveController; 
    umin    = nothing, umax    = nothing,
    Δumin   = nothing, Δumax   = nothing,
    ymin    = nothing, ymax    = nothing,
    x̂min    = nothing, x̂max    = nothing,
    c_umin  = nothing, c_umax  = nothing,
    c_Δumin = nothing, c_Δumax = nothing,
    c_ymin  = nothing, c_ymax  = nothing,
    c_x̂min  = nothing, c_x̂max  = nothing,
    Umin    = nothing, Umax    = nothing,
    ΔUmin   = nothing, ΔUmax   = nothing,
    Ymin    = nothing, Ymax    = nothing,
    C_umax  = nothing, C_umin  = nothing,
    C_Δumax = nothing, C_Δumin = nothing,
    C_ymax  = nothing, C_ymin  = nothing,
    # ------------ will be deleted in the future ---------------
    ŷmin    = nothing, ŷmax    = nothing,
    c_ŷmin  = nothing, c_ŷmax  = nothing,
    # ----------------------------------------------------------
)
    # ----- these 4 `if`s will be deleted in the future --------
    if !isnothing(ŷmin)
        Base.depwarn("keyword arg ŷmin is deprecated, use ymin instead", :setconstraint!)
        ymin = ŷmin
    end
    if !isnothing(ŷmax)
        Base.depwarn("keyword arg ŷmax is deprecated, use ymax instead", :setconstraint!)
        ymax = ŷmax
    end
    if !isnothing(c_ŷmin)
        Base.depwarn("keyword arg ŷmin is deprecated, use ymin instead", :setconstraint!)
        c_ymin = c_ŷmin
    end
    if !isnothing(c_ŷmax)
        Base.depwarn("keyword arg ŷmax is deprecated, use ymax instead", :setconstraint!)
        c_ymax = c_ŷmax
    end
    # ----------------------------------------------------------
    model, con, optim = mpc.estim.model, mpc.con, mpc.optim
    nu, ny, nx̂, Hp, Hc = model.nu, model.ny, mpc.estim.nx̂, mpc.Hp, mpc.Hc
    notSolvedYet = (termination_status(optim) == OPTIMIZE_NOT_CALLED)
    C, E, ex̂ = mpc.C, mpc.Ẽ[:, 1:nu*Hc], mpc.con.ẽx̂[:, 1:nu*Hc]
    isnothing(Umin)     && !isnothing(umin)     && (Umin    = repeat(umin,    Hp))
    isnothing(Umax)     && !isnothing(umax)     && (Umax    = repeat(umax,    Hp))
    isnothing(ΔUmin)    && !isnothing(Δumin)    && (ΔUmin   = repeat(Δumin,   Hc))
    isnothing(ΔUmax)    && !isnothing(Δumax)    && (ΔUmax   = repeat(Δumax,   Hc))
    isnothing(Ymin)     && !isnothing(ymin)     && (Ymin    = repeat(ymin,    Hp))
    isnothing(Ymax)     && !isnothing(ymax)     && (Ymax    = repeat(ymax,    Hp))
    isnothing(C_umin)   && !isnothing(c_umin)   && (C_umin  = repeat(c_umin,  Hp))
    isnothing(C_umax)   && !isnothing(c_umax)   && (C_umax  = repeat(c_umax,  Hp))
    isnothing(C_Δumin)  && !isnothing(c_Δumin)  && (C_Δumin = repeat(c_Δumin, Hc))
    isnothing(C_Δumax)  && !isnothing(c_Δumax)  && (C_Δumax = repeat(c_Δumax, Hc))
    isnothing(C_ymin)   && !isnothing(c_ymin)   && (C_ymin  = repeat(c_ymin,  Hp))
    isnothing(C_ymax)   && !isnothing(c_ymax)   && (C_ymax  = repeat(c_ymax,  Hp))
    if !all(isnothing.([C_umin, C_umax, C_Δumin, C_Δumax, C_ymin, C_ymax, c_x̂min, c_x̂max]))
        !isinf(C) || throw(ArgumentError("Slack variable weight Cwt must be finite to set softness parameters"))
        notSolvedYet || error("Cannot set softness parameters after calling moveinput!")
    end
    if !isnothing(Umin)
        size(Umin)   == (nu*Hp,) || throw(ArgumentError("Umin size must be $((nu*Hp,))"))
        con.Umin[:] = Umin
    end
    if !isnothing(Umax)
        size(Umax)   == (nu*Hp,) || throw(ArgumentError("Umax size must be $((nu*Hp,))"))
        con.Umax[:] = Umax
    end
    if !isnothing(ΔUmin)
        size(ΔUmin)  == (nu*Hc,) || throw(ArgumentError("ΔUmin size must be $((nu*Hc,))"))
        con.ΔŨmin[1:nu*Hc] = ΔUmin
    end
    if !isnothing(ΔUmax)
        size(ΔUmax)  == (nu*Hc,) || throw(ArgumentError("ΔUmax size must be $((nu*Hc,))"))
        con.ΔŨmax[1:nu*Hc] = ΔUmax
    end
    if !isnothing(Ymin)
        size(Ymin)   == (ny*Hp,) || throw(ArgumentError("Ymin size must be $((ny*Hp,))"))
        con.Ymin[:] = Ymin
    end
    if !isnothing(Ymax)
        size(Ymax)   == (ny*Hp,) || throw(ArgumentError("Ymax size must be $((ny*Hp,))"))
        con.Ymax[:] = Ymax
    end
    if !isnothing(x̂min)
        size(x̂min)   == (nx̂,) || throw(ArgumentError("x̂min size must be $((nx̂,))"))
        con.x̂min[:] = x̂min
    end
    if !isnothing(x̂max)
        size(x̂max)   == (nx̂,) || throw(ArgumentError("x̂max size must be $((nx̂,))"))
        con.x̂max[:] = x̂max
    end
    if !isnothing(C_umin)
        size(C_umin) == (nu*Hp,) || throw(ArgumentError("C_umin size must be $((nu*Hp,))"))
        any(C_umin .< 0) && error("C_umin weights should be non-negative")
        con.A_Umin[:, end] = -C_umin
    end
    if !isnothing(C_umax)
        size(C_umax) == (nu*Hp,) || throw(ArgumentError("C_umax size must be $((nu*Hp,))"))
        any(C_umax .< 0) && error("C_umax weights should be non-negative")
        con.A_Umax[:, end] = -C_umax
    end
    if !isnothing(C_Δumin)
        size(C_Δumin) == (nu*Hc,) || throw(ArgumentError("C_Δumin size must be $((nu*Hc,))"))
        any(C_Δumin .< 0) && error("C_Δumin weights should be non-negative")
        con.A_ΔŨmin[1:end-1, end] = -C_Δumin 
    end
    if !isnothing(C_Δumax)
        size(C_Δumax) == (nu*Hc,) || throw(ArgumentError("C_Δumax size must be $((nu*Hc,))"))
        any(C_Δumax .< 0) && error("C_Δumax weights should be non-negative")
        con.A_ΔŨmax[1:end-1, end] = -C_Δumax
    end
    if !isnothing(C_ymin)
        size(C_ymin) == (ny*Hp,) || throw(ArgumentError("C_ymin size must be $((ny*Hp,))"))
        any(C_ymin .< 0) && error("C_ymin weights should be non-negative")
        con.C_ymin[:] = C_ymin
        A_Ymin ,_ = relaxŶ(model, C, con.C_ymin, con.C_ymax, E)
        con.A_Ymin[:] = A_Ymin
    end
    if !isnothing(C_ymax)
        size(C_ymax) == (ny*Hp,) || throw(ArgumentError("C_ymax size must be $((ny*Hp,))"))
        any(C_ymax .< 0) && error("C_ymax weights should be non-negative")
        con.C_ymax[:] = C_ymax
        _, A_Ymax = relaxŶ(model, C, con.C_ymin, con.C_ymax, E)
        con.A_Ymax[:] = A_Ymax
    end
    if !isnothing(c_x̂min)
        size(c_x̂min) == (nx̂,) || throw(ArgumentError("c_x̂min size must be $((nx̂,))"))
        any(c_x̂min .< 0) && error("c_x̂min weights should be non-negative")
        con.c_x̂min[:] = c_x̂min
        A_x̂min ,_ = relaxterminal(model, C, con.c_x̂min, con.c_x̂max, ex̂)
        con.A_x̂min[:] = A_x̂min
    end
    if !isnothing(c_x̂max)
        size(c_x̂max) == (nx̂,) || throw(ArgumentError("c_x̂max size must be $((nx̂,))"))
        any(c_x̂max .< 0) && error("c_x̂max weights should be non-negative")
        con.c_x̂max[:] = c_x̂max
        _, A_x̂max = relaxterminal(model, C, con.c_x̂min, con.c_x̂max, ex̂)
        con.A_x̂max[:] = A_x̂max
    end
    i_Umin,  i_Umax  = .!isinf.(con.Umin),  .!isinf.(con.Umax)
    i_ΔŨmin, i_ΔŨmax = .!isinf.(con.ΔŨmin), .!isinf.(con.ΔŨmin)
    i_Ymin,  i_Ymax  = .!isinf.(con.Ymin),  .!isinf.(con.Ymax)
    i_x̂min,  i_x̂max  = .!isinf.(con.x̂min),  .!isinf.(con.x̂max)
    if notSolvedYet
        con.i_b[:], con.i_g[:], con.A[:] = init_matconstraint(model,
            i_Umin, i_Umax, i_ΔŨmin, i_ΔŨmax, 
            i_Ymin, i_Ymax, i_x̂min, i_x̂max,
            con.A_Umin, con.A_Umax, con.A_ΔŨmin, con.A_ΔŨmax, 
            con.A_Ymin, con.A_Ymax, con.A_x̂min, con.A_x̂max
        )
        A = con.A[con.i_b, :]
        b = con.b[con.i_b]
        ΔŨvar = mpc.optim[:ΔŨvar]
        delete(mpc.optim, mpc.optim[:linconstraint])
        unregister(mpc.optim, :linconstraint)
        @constraint(mpc.optim, linconstraint, A*ΔŨvar .≤ b)
        setnonlincon!(mpc, model)
    else
        i_b, i_g = init_matconstraint(model, 
            i_Umin, i_Umax, i_ΔŨmin, i_ΔŨmax, 
            i_Ymin, i_Ymax, i_x̂min, i_x̂max
        )
        if i_b ≠ con.i_b || i_g ≠ con.i_g
            error("Cannot modify ±Inf constraints after calling moveinput!")
        end
    end
    return mpc
end



"By default, there is no nonlinear constraint, thus do nothing."
setnonlincon!(::PredictiveController, ::SimModel) = nothing

@doc raw"""
    moveinput!(mpc::PredictiveController, ry=mpc.estim.model.yop, d=[]; <keyword args>) -> u

Compute the optimal manipulated input value `u` for the current control period.

Solve the optimization problem of `mpc` [`PredictiveController`](@ref) and return the
results ``\mathbf{u}(k)``. Following the receding horizon principle, the algorithm discards
the optimal future manipulated inputs ``\mathbf{u}(k+1), \mathbf{u}(k+2), ...`` Note that
the method mutates `mpc` internal data but it does not modifies `mpc.estim` states. Call
[`updatestate!(mpc, u, ym, d)`](@ref) to update `mpc` state estimates.

Calling a [`PredictiveController`](@ref) object calls this method.

See also [`LinMPC`](@ref), [`ExplicitMPC`](@ref), [`NonLinMPC`](@ref).

# Arguments
- `mpc::PredictiveController` : solve optimization problem of `mpc`.
- `ry=mpc.estim.model.yop` : current output setpoints ``\mathbf{r_y}(k)``.
- `d=[]` : current measured disturbances ``\mathbf{d}(k)``.
- `D̂=repeat(d, mpc.Hp)` : predicted measured disturbances ``\mathbf{D̂}``, constant in the
  future by default or ``\mathbf{d̂}(k+j)=\mathbf{d}(k)`` for ``j=1`` to ``H_p``.
- `R̂y=repeat(ry, mpc.Hp)` : predicted output setpoints ``\mathbf{R̂_y}``, constant in the
  future by default or ``\mathbf{r̂_y}(k+j)=\mathbf{r_y}(k)`` for ``j=1`` to ``H_p``.
- `R̂u=repeat(mpc.estim.model.uop, mpc.Hp)` : predicted manipulated input setpoints, constant
  in the future by default or ``\mathbf{r̂_u}(k+j)=\mathbf{u_{op}}`` for ``j=0`` to ``H_p-1``.
- `ym=nothing` : current measured outputs ``\mathbf{y^m}(k)``, only required if `mpc.estim` 
   is an [`InternalModel`](@ref).

# Examples
```jldoctest
julia> mpc = LinMPC(LinModel(tf(5, [2, 1]), 3), Nwt=[0], Hp=1000, Hc=1);

julia> ry = [5]; u = moveinput!(mpc, ry); round.(u, digits=3)
1-element Vector{Float64}:
 1.0
```
"""
function moveinput!(
    mpc::PredictiveController, 
    ry::Vector = mpc.estim.model.yop, 
    d ::Vector = empty(mpc.estim.x̂);
    D̂ ::Vector = repeat(d,  mpc.Hp),
    R̂y::Vector = repeat(ry, mpc.Hp),
    R̂u::Vector = mpc.noR̂u ? empty(mpc.estim.x̂) : repeat(mpc.estim.model.uop, mpc.Hp),
    ym::Union{Vector, Nothing} = nothing
)
    validate_setpointdist(mpc, ry, d, D̂, R̂y, R̂u)
    initpred!(mpc, mpc.estim.model, d, ym, D̂, R̂y, R̂u)
    linconstraint!(mpc, mpc.estim.model)
    ΔŨ = optim_objective!(mpc)
    Δu = ΔŨ[1:mpc.estim.model.nu] # receding horizon principle: only Δu(k) is used (1st one)
    u = mpc.estim.lastu0 + mpc.estim.model.uop + Δu
    return u
end

@doc raw"""
    getinfo(mpc::PredictiveController) -> info

Get additional information about `mpc` controller optimum to ease troubleshooting.

The function should be called after calling [`moveinput!`](@ref). It returns the dictionary
`info` with the following fields:

- `:ΔU`  : optimal manipulated input increments over `Hc`, ``\mathbf{ΔU}``.
- `:ϵ`   : optimal slack variable, ``ϵ``.
- `:J`   : objective value optimum, ``J``.
- `:U`   : optimal manipulated inputs over `Hp`, ``\mathbf{U}``.
- `:u`   : current optimal manipulated input, ``\mathbf{u}(k)``.
- `:d`   : current measured disturbance, ``\mathbf{d}(k)``.
- `:D̂`   : predicted measured disturbances over `Hp`, ``\mathbf{D̂}``.
- `:ŷ`   : current estimated output, ``\mathbf{ŷ}(k)``.
- `:Ŷ`   : optimal predicted outputs over `Hp`, ``\mathbf{Ŷ}``.
- `:x̂end`: optimal terminal states, ``\mathbf{x̂}_{k-1}(k+H_p)``.
- `:Ŷs`  : predicted stochastic output over `Hp` of [`InternalModel`](@ref), ``\mathbf{Ŷ_s}``.
- `:R̂y`  : predicted output setpoint over `Hp`, ``\mathbf{R̂_y}``.
- `:R̂u`  : predicted manipulated input setpoint over `Hp`, ``\mathbf{R̂_u}``.

For [`LinMPC`](@ref) and [`NonLinMPC`](@ref), the field `:sol` also contains the optimizer
solution summary that can be printed. Lastly, the optimal economic cost `:JE` is also
available for [`NonLinMPC`](@ref).

# Examples
```jldoctest
julia> mpc = LinMPC(LinModel(tf(5, [2, 1]), 3), Nwt=[0], Hp=1, Hc=1);

julia> u = moveinput!(mpc, [10]);

julia> round.(getinfo(mpc)[:Ŷ], digits=3)
1-element Vector{Float64}:
 10.0
```
"""
function getinfo(mpc::PredictiveController)
    info = Dict{Symbol, InfoDictType}()
    Ŷ, x̂    = similar(mpc.Ŷop), similar(mpc.estim.x̂)
    Ŷ, x̂end = predict!(Ŷ, x̂, mpc, mpc.estim.model, mpc.ΔŨ)
    info[:ΔU]   = mpc.ΔŨ[1:mpc.Hc*mpc.estim.model.nu]
    info[:ϵ]    = isinf(mpc.C) ? NaN : mpc.ΔŨ[end]
    info[:J]    = obj_nonlinprog(mpc, mpc.estim.model, Ŷ, mpc.ΔŨ) + mpc.p[]
    info[:U]    = mpc.S̃*mpc.ΔŨ + mpc.T*(mpc.estim.lastu0 + mpc.estim.model.uop)
    info[:u]    = info[:U][1:mpc.estim.model.nu]
    info[:d]    = mpc.d0 + mpc.estim.model.dop
    info[:D̂]    = mpc.D̂0 + mpc.Dop
    info[:ŷ]    = mpc.ŷ
    info[:Ŷ]    = Ŷ
    info[:x̂end] = x̂end
    info[:Ŷs]  = mpc.Ŷop - repeat(mpc.estim.model.yop, mpc.Hp) # Ŷop = Ŷs + Yop
    info[:R̂y]  = mpc.R̂y
    info[:R̂u]  = mpc.R̂u
    info = addinfo!(info, mpc)
    return info
end

"""
    addinfo!(info, mpc::PredictiveController) -> info

By default, add the solution summary `:sol` that can be printed to `info`.
"""
function addinfo!(info, mpc::PredictiveController)
    info[:sol] = solution_summary(mpc.optim, verbose=true)
    return info
end


"""
    setstate!(mpc::PredictiveController, x̂)

Set the estimate at `mpc.estim.x̂`.
"""
setstate!(mpc::PredictiveController, x̂) = (setstate!(mpc.estim, x̂); return mpc)

@doc raw"""
    initstate!(mpc::PredictiveController, u, ym, d=[]) -> x̂

Init the states of `mpc.estim` [`StateEstimator`](@ref) and warm start `mpc.ΔŨ` at zero.
"""
function initstate!(mpc::PredictiveController, u, ym, d=empty(mpc.estim.x̂))
    mpc.ΔŨ .= 0
    return initstate!(mpc.estim, u, ym, d)
end


"""
    updatestate!(mpc::PredictiveController, u, ym, d=[]) -> x̂

Call [`updatestate!`](@ref) on `mpc.estim` [`StateEstimator`](@ref).
"""
updatestate!(mpc::PredictiveController, u, ym, d=empty(mpc.estim.x̂)) = updatestate!(mpc.estim,u,ym,d)
updatestate!(::PredictiveController, _ ) = throw(ArgumentError("missing measured outputs ym"))

"""
    default_Hp(model::LinModel, Hp)

Estimate the default prediction horizon `Hp` with a security margin for [`LinModel`](@ref).
"""
function default_Hp(model::LinModel, Hp)
    # TODO: also check for settling time (poles)
    # TODO: also check for non minimum phase systems (zeros)
    # TODO: replace sum with max delay between all the I/O
    # TODO: use this nk value for default N value in sim!
    poles = eigvals(model.A)
    # atol=1e-3 to overestimate the number of delays : for closed-loop stability, it is
    # better to overestimate the default value of Hp, as a security margin.
    nk = sum(isapprox.(abs.(poles), 0.0, atol=1e-3)) # number of delays
    if isnothing(Hp)
        Hp = DEFAULT_HP0 + nk
    end
    if Hp ≤ nk
        @warn("prediction horizon Hp ($Hp) ≤ estimated number of delays in model "*
              "($nk), the closed-loop system may be unstable or zero-gain (unresponsive)")
    end
    return Hp
end

"""
    default_Hp(model::SimModel, Hp)

Throw an error if `isnothing(Hp)` when model is not a [`LinModel`](@ref).
"""
function default_Hp(::SimModel, Hp)
    if isnothing(Hp)
        # ------------ will be deleted in the future ------------------------------------
        Base.depwarn("Hp=nothing is deprecated for NonLinModel, explicitly specify an "*
                     "integer value", :NonLinMPC)
        Hp = DEFAULT_HP0
        # ------------- and replaced by this -------------------------------------------
        # throw(ArgumentError("Prediction horizon Hp must be explicitly specified if "*
        #                     "model is not a LinModel."))
        # Hp = 0
        # -----------------------------------------------------------------------------
    end
    return Hp
end

function validate_setpointdist(mpc::PredictiveController, ry, d, D̂, R̂y, R̂u)
    ny, nd, nu, Hp = mpc.estim.model.ny, mpc.estim.model.nd, mpc.estim.model.nu, mpc.Hp
    size(ry) ≠ (ny,)    && throw(ArgumentError("ry size $(size(ry)) ≠ output size ($ny,)"))
    size(d)  ≠ (nd,)    && throw(ArgumentError("d size $(size(d)) ≠ measured dist. size ($nd,)"))
    size(D̂)  ≠ (nd*Hp,) && throw(ArgumentError("D̂ size $(size(D̂)) ≠ measured dist. size × Hp ($(nd*Hp),)"))
    size(R̂y) ≠ (ny*Hp,) && throw(ArgumentError("R̂y size $(size(R̂y)) ≠ output size × Hp ($(ny*Hp),)"))
    if ~mpc.noR̂u
        size(R̂u) ≠ (nu*Hp,) && throw(ArgumentError("R̂u size $(size(R̂u)) ≠ manip. input size × Hp ($(nu*Hp),)"))
    end
end

@doc raw"""
    initpred!(mpc, model::LinModel, d, ym, D̂, R̂y, R̂u)

Init linear model prediction matrices `F, q̃, p` and current estimated output `ŷ`.

See [`init_predmat`](@ref) and [`init_quadprog`](@ref) for the definition of the matrices.
"""
function initpred!(mpc::PredictiveController, model::LinModel, d, ym, D̂, R̂y, R̂u)
    mpc.ŷ[:] = evalŷ(mpc.estim, ym, d)
    predictstoch!(mpc, mpc.estim, d, ym) # init mpc.Ŷop for InternalModel
    mpc.F[:]  = mpc.K  * mpc.estim.x̂  + mpc.V  * mpc.estim.lastu0 + mpc.Ŷop
    if model.nd ≠ 0
        mpc.d0[:], mpc.D̂0[:] = d - model.dop, D̂ - mpc.Dop
        mpc.D̂E[:] = [d; D̂]
        mpc.F[:]  = mpc.F  + mpc.G  * mpc.d0 + mpc.J  * mpc.D̂0
    end
    mpc.R̂y[:] = R̂y
    Ẑ = mpc.F - mpc.R̂y
    mpc.q̃[:] = 2(mpc.M_Hp*mpc.Ẽ)'*Ẑ
    mpc.p[]  = dot(Ẑ, mpc.M_Hp, Ẑ)
    if ~mpc.noR̂u
        mpc.R̂u[:] = R̂u
        lastu = mpc.estim.lastu0 + model.uop
        Ẑ = mpc.T*lastu - mpc.R̂u
        mpc.q̃[:] = mpc.q̃ + 2(mpc.L_Hp*mpc.S̃)'*Ẑ
        mpc.p[]  = mpc.p[] + dot(Ẑ, mpc.L_Hp, Ẑ)
    end
    return nothing
end

@doc raw"""
    initpred!(mpc::PredictiveController, model::SimModel, d, ym, D̂, R̂y, R̂u)

Init `ŷ, Ŷop, d0, D̂0` matrices when model is not a [`LinModel`](@ref).

`d0` and `D̂0` are the measured disturbances and its predictions without the operating points
``\mathbf{d_{op}}``. The vector `Ŷop` is kept unchanged if `mpc.estim` is not an
[`InternalModel`](@ref).
"""
function initpred!(mpc::PredictiveController, model::SimModel, d, ym, D̂, R̂y, R̂u)
    mpc.ŷ[:] = evalŷ(mpc.estim, ym, d)
    predictstoch!(mpc, mpc.estim, d, ym) # init mpc.Ŷop for InternalModel
    if model.nd ≠ 0
        mpc.d0[:], mpc.D̂0[:] = d - model.dop, D̂ - mpc.Dop
        mpc.D̂E[:] = [d; D̂]
    end
    mpc.R̂y[:] = R̂y
    if ~mpc.noR̂u
        mpc.R̂u[:] = R̂u
    end
    return nothing
end

@doc raw"""
    predictstoch!(mpc::PredictiveController, estim::InternalModel, x̂s, d, ym)

Init `Ŷop` vector when if `estim` is an [`InternalModel`](@ref).

The vector combines the output operating points and the stochastic predictions:
``\mathbf{Ŷ_{op} = Ŷ_{s} + Y_{op}}`` (both values are constant between the nonlinear 
programming iterations).
"""
function predictstoch!(mpc::PredictiveController, estim::InternalModel, d, ym)
    isnothing(ym) && error("Predictive controllers with InternalModel need the measured "*
                           "outputs ym in keyword argument to compute control actions u")
    ŷd = h(estim.model, estim.x̂d, d - estim.model.dop) + estim.model.yop 
    ŷs = zeros(estim.model.ny)
    ŷs[estim.i_ym] = ym - ŷd[estim.i_ym]  # ŷs=0 for unmeasured outputs
    Ŷs = mpc.Ks*mpc.estim.x̂s + mpc.Ps*ŷs
    mpc.Ŷop[:] = Ŷs + repeat(estim.model.yop, mpc.Hp)
    return nothing
end
"Separate stochastic predictions are not needed if `estim` is not [`InternalModel`](@ref)."
predictstoch!(::PredictiveController, ::StateEstimator, _ , _ ) = nothing

@doc raw"""
    predict!(Ŷ, x̂, mpc::PredictiveController, model::LinModel, ΔŨ) -> Ŷ, x̂end

Compute the predictions `Ŷ` and terminal states `x̂end` if model is a [`LinModel`](@ref).

The method mutates `Ŷ` and `x̂` vector arguments. The `x̂end` vector is used for
the terminal constraints applied on ``\mathbf{x̂}_{k-1}(k+H_p)``.
"""
function predict!(Ŷ, x̂, mpc::PredictiveController, ::LinModel, ΔŨ::Vector{T}) where {T<:Real}
     # in-place operations to reduce allocations :
    Ŷ[:] = mul!(Ŷ, mpc.Ẽ, ΔŨ) + mpc.F
    x̂[:] = mul!(x̂, mpc.con.ẽx̂, ΔŨ) + mpc.con.fx̂
    x̂end = x̂
    return Ŷ, x̂end
end

@doc raw"""
    predict!(Ŷ, x̂, mpc::PredictiveController, model::SimModel, ΔŨ) -> Ŷ, x̂end

Compute both vectors if `model` is not a [`LinModel`](@ref).
"""
function predict!(Ŷ, x̂, mpc::PredictiveController, model::SimModel, ΔŨ::Vector{T}) where {T<:Real}
    nu, ny, nd, Hp, Hc = model.nu, model.ny, model.nd, mpc.Hp, mpc.Hc
    u0::Vector{T} = copy(mpc.estim.lastu0)
    d0::Vector{T} = copy(mpc.d0)
    x̂[:] = mpc.estim.x̂
    for j=1:Hp
        if j ≤ Hc
            u0[:] = u0 + ΔŨ[(1 + nu*(j-1)):(nu*j)]
        end
        x̂[:]  = f̂(mpc.estim, model, x̂, u0, d0)
        d0[:] = mpc.D̂0[(1 + nd*(j-1)):(nd*j)]
        Ŷ[(1 + ny*(j-1)):(ny*j)] = ĥ(mpc.estim, model, x̂, d0)
    end
    Ŷ[:] = Ŷ + mpc.Ŷop # Ŷop = Ŷs + Yop, and Ŷs=0 if mpc.estim is not an InternalModel
    x̂end = x̂
    return Ŷ, x̂end
end

@doc raw"""
    linconstraint!(mpc::PredictiveController, model::LinModel)

Set `b` vector for the linear model inequality constraints (``\mathbf{A ΔŨ ≤ b}``).

Also init ``\mathbf{f_x̂}`` vector for the terminal constraints, see [`init_predmat`](@ref).
"""
function linconstraint!(mpc::PredictiveController, model::LinModel)
    mpc.con.fx̂[:] = mpc.con.kx̂ * mpc.estim.x̂  + mpc.con.vx̂ * mpc.estim.lastu0
    if model.nd ≠ 0
        mpc.con.fx̂[:] = mpc.con.fx̂ + mpc.con.gx̂ * mpc.d0 + mpc.con.jx̂ * mpc.D̂0
    end
    lastu = mpc.estim.lastu0 + model.uop
    mpc.con.b[:] = [
        -mpc.con.Umin + mpc.T*lastu
        +mpc.con.Umax - mpc.T*lastu
        -mpc.con.ΔŨmin
        +mpc.con.ΔŨmax 
        -mpc.con.Ymin + mpc.F
        +mpc.con.Ymax - mpc.F
        -mpc.con.x̂min + mpc.con.fx̂
        +mpc.con.x̂max - mpc.con.fx̂
    ]
    lincon::LinConVector = mpc.optim[:linconstraint]
    set_normalized_rhs.(lincon, mpc.con.b[mpc.con.i_b])
end

"Set `b` excluding predicted output constraints when `model` is not a [`LinModel`](@ref)."
function linconstraint!(mpc::PredictiveController, model::SimModel)
    lastu = mpc.estim.lastu0 + model.uop
    mpc.con.b[:] = [
        -mpc.con.Umin + mpc.T*lastu
        +mpc.con.Umax - mpc.T*lastu
        -mpc.con.ΔŨmin
        +mpc.con.ΔŨmax 
    ]
    lincon::LinConVector = mpc.optim[:linconstraint]
    set_normalized_rhs.(lincon, mpc.con.b[mpc.con.i_b])
end

"""
    optim_objective!(mpc::PredictiveController)

Optimize the objective function ``J`` of `mpc` controller and return the solution `ΔŨ`.
"""
function optim_objective!(mpc::PredictiveController)
    optim = mpc.optim
    model = mpc.estim.model
    ΔŨvar::Vector{VariableRef} = optim[:ΔŨvar]
    lastΔŨ = mpc.ΔŨ
    # initial ΔŨ (warm-start): [Δu_{k-1}(k); Δu_{k-1}(k+1); ... ; 0_{nu × 1}; ϵ_{k-1}]
    ϵ0  = !isinf(mpc.C) ? [lastΔŨ[end]] : empty(mpc.ΔŨ)
    ΔŨ0 = [lastΔŨ[(model.nu+1):(mpc.Hc*model.nu)]; zeros(model.nu); ϵ0]
    set_start_value.(ΔŨvar, ΔŨ0)
    set_objective_linear_coef!(mpc, ΔŨvar)
    try
        optimize!(optim)
    catch err
        if isa(err, MOI.UnsupportedAttribute{MOI.VariablePrimalStart})
            # reset_optimizer to unset warm-start, set_start_value.(nothing) seems buggy
            MOIU.reset_optimizer(optim)
            optimize!(optim)
        else
            rethrow(err)
        end
    end
    status = termination_status(optim)
    ΔŨcurr, ΔŨlast = value.(ΔŨvar), ΔŨ0
    if !(status == OPTIMAL || status == LOCALLY_SOLVED)
        if isfatal(status)
            @error("MPC terminated without solution: returning last solution shifted", 
                   status, ΔŨcurr, ΔŨlast)
        else
            @warn("MPC termination status not OPTIMAL or LOCALLY_SOLVED: keeping "*
                  "solution anyway", status, ΔŨcurr, ΔŨlast)
        end
        @debug solution_summary(optim, verbose=true)
    end
    mpc.ΔŨ[:] = isfatal(status) ? ΔŨlast : ΔŨcurr
    return mpc.ΔŨ
end

"By default, no need to modify the objective function."
set_objective_linear_coef!(::PredictiveController, _ ) = nothing

@doc raw"""
    init_ΔUtoU(nu, Hp) -> S, T

Init manipulated input increments to inputs conversion matrices.

The conversion from the input increments ``\mathbf{ΔU}`` to manipulated inputs over ``H_p`` 
are calculated by:
```math
\mathbf{U} = \mathbf{S} \mathbf{ΔU} + \mathbf{T} \mathbf{u}(k-1) \\
```
"""
function init_ΔUtoU(nu, Hp, Hc)
    S_Hc = LowerTriangular(repeat(I(nu), Hc, Hc))
    S = [S_Hc; repeat(I(nu), Hp - Hc, Hc)]
    T = repeat(I(nu), Hp)
    return S, T
end


@doc raw"""
    init_predmat(estim, ::LinModel, Hp, Hc) -> E, G, J, K, V, ex̂, fx̂, gx̂, jx̂, kx̂, vx̂

Construct the prediction matrices for [`LinModel`](@ref) `model`.

The linear model predictions are evaluated by :
```math
\begin{aligned}
    \mathbf{Ŷ} &= \mathbf{E ΔU} + \mathbf{G d}(k) + \mathbf{J D̂} 
                                + \mathbf{K x̂}_{k-1}(k) + \mathbf{V u}(k-1) \\
               &= \mathbf{E ΔU} + \mathbf{F}
\end{aligned}
```
where predicted outputs ``\mathbf{Ŷ}``, stochastic outputs ``\mathbf{Ŷ_s}``, and measured
disturbances ``\mathbf{D̂}`` are from ``k + 1`` to ``k + H_p``. Input increments 
``\mathbf{ΔU}`` are from ``k`` to ``k + H_c - 1``. The vector ``\mathbf{x̂}_{k-1}(k)`` is the
state estimated at the last control period. The method also computes similar matrices but 
for the predicted terminal states at ``k+H_p``:
```math
\begin{aligned}
    \mathbf{x̂}_{k-1}(k+H_p) 
            &= \mathbf{e_x̂ ΔU} + \mathbf{g_x̂ d}(k) + \mathbf{j_x̂ D̂} 
                               + \mathbf{k_x̂ x̂}_{k-1}(k) + \mathbf{v_x̂ u}(k-1) \\
            &= \mathbf{e_x̂ ΔU} + \mathbf{f_x̂}
\end{aligned}
```
Operating points on ``\mathbf{u}``, ``\mathbf{d}`` and ``\mathbf{y}`` are omitted in above
equations.

# Extended Help
Using the augmented matrices ``\mathbf{Â, B̂_u, Ĉ, B̂_d, D̂_d}`` in `estim` and the function
``\mathbf{W}(j) = \mathbf{Ĉ} ( ∑_{i=0}^j \mathbf{Â}^i ) \mathbf{B̂_u}``, the prediction 
matrices are computed by :
```math
\begin{aligned}
\mathbf{E} &= \begin{bmatrix}
    \mathbf{W}(0)      & \mathbf{0}        & \cdots & \mathbf{0}              \\
    \mathbf{W}(1)      & \mathbf{W}(0)     & \cdots & \mathbf{0}              \\
    \vdots             & \vdots            & \ddots & \vdots                  \\
    \mathbf{W}(H_p-1)  & \mathbf{W}(H_p-2) & \cdots & \mathbf{W}(H_p-H_c+1)   \end{bmatrix} \\
\mathbf{G} &= \begin{bmatrix}
    \mathbf{Ĉ}\mathbf{Â}^{0} \mathbf{B̂_d}     \\ 
    \mathbf{Ĉ}\mathbf{Â}^{1} \mathbf{B̂_d}     \\ 
    \vdots                                    \\
    \mathbf{Ĉ}\mathbf{Â}^{H_p-1} \mathbf{B̂_d} \end{bmatrix} \\
\mathbf{J} &= \begin{bmatrix}
    \mathbf{D̂_d}                              & \mathbf{0}                                & \cdots & \mathbf{0}   \\ 
    \mathbf{Ĉ}\mathbf{Â}^{0} \mathbf{B̂_d}     & \mathbf{D̂_d}                              & \cdots & \mathbf{0}   \\ 
    \vdots                                    & \vdots                                    & \ddots & \vdots       \\
    \mathbf{Ĉ}\mathbf{Â}^{H_p-2} \mathbf{B̂_d} & \mathbf{Ĉ}\mathbf{Â}^{H_p-3} \mathbf{B̂_d} & \cdots & \mathbf{D̂_d} \end{bmatrix} \\
\mathbf{K} &= \begin{bmatrix}
    \mathbf{Ĉ}\mathbf{Â}^{1}      \\
    \mathbf{Ĉ}\mathbf{Â}^{2}      \\
    \vdots                        \\
    \mathbf{Ĉ}\mathbf{Â}^{H_p}    \end{bmatrix} \\
\mathbf{V} &= \begin{bmatrix}
    \mathbf{W}(0)        \\
    \mathbf{W}(1)        \\
    \vdots               \\
    \mathbf{W}(H_p-1)    \end{bmatrix}
\end{aligned}
```
For the terminal constraints, the matrices are computed with the function
``\mathbf{w_x̂}(j) = ( ∑_{i=0}^j \mathbf{Â}^i ) \mathbf{B̂_u}`` and:
```math
\begin{aligned}
\mathbf{e_x̂} &= \begin{bmatrix} \mathbf{w_x̂}(H_p-1) & \mathbf{w_x̂}(H_p-2) & \cdots & \mathbf{w_x̂}(H_p-H_c+1) \end{bmatrix} \\
\mathbf{g_x̂} &= \mathbf{Â}^{H_p-1} \mathbf{B̂_d} \\
\mathbf{j_x̂} &= \begin{bmatrix} \mathbf{Â}^{H_p-2} \mathbf{B̂_d} & \mathbf{Â}^{H_p-3} \mathbf{B̂_d} & \cdots & \mathbf{0} \end{bmatrix} \\
\mathbf{k_x̂} &= \mathbf{Â}^{H_p} \\
\mathbf{v_x̂} &= \mathbf{w_x̂}(H_p-1)
\end{aligned}
```
"""
function init_predmat(estim::StateEstimator, model::LinModel, Hp, Hc)
    Â, B̂u, Ĉ, B̂d, D̂d = estim.Â, estim.B̂u, estim.Ĉ, estim.B̂d, estim.D̂d
    nu, nx̂, ny, nd = model.nu, estim.nx̂, model.ny, model.nd
    # --- pre-compute matrix powers ---
    # Apow 3D array : Apow[:,:,1] = A^0, Apow[:,:,2] = A^1, ... , Apow[:,:,Hp+1] = A^Hp
    Âpow = Array{Float64}(undef, size(Â,1), size(Â,2), Hp+1)
    Âpow[:,:,1] = I(nx̂)
    for j=2:Hp+1
        Âpow[:,:,j] = Âpow[:,:,j-1]*Â
    end
    # Apow_csum 3D array : Apow_csum[:,:,1] = A^0, Apow_csum[:,:,2] = A^1 + A^0, ...
    Âpow_csum  = cumsum(Âpow, dims=3)
    # helper function to improve code clarity and be similar to eqs. in docstring:
    getpower(array3D, power) = array3D[:,:, power+1]
    # --- state estimates x̂ ---
    kx̂ = getpower(Âpow, Hp)
    K  = Matrix{Float64}(undef, Hp*ny, nx̂)
    for j=1:Hp
        iRow = (1:ny) .+ ny*(j-1)
        K[iRow,:] = Ĉ*getpower(Âpow, j)
    end    
    # --- manipulated inputs u ---
    vx̂ = getpower(Âpow_csum, Hp-1)*B̂u
    V  = Matrix{Float64}(undef, Hp*ny, nu)
    for j=1:Hp
        iRow = (1:ny) .+ ny*(j-1)
        V[iRow,:] = Ĉ*getpower(Âpow_csum, j-1)*B̂u
    end
    ex̂ = Matrix{Float64}(undef, nx̂, Hc*nu)
    E  = zeros(Hp*ny, Hc*nu) 
    for j=1:Hc # truncated with control horizon
        iRow = (ny*(j-1)+1):(ny*Hp)
        iCol = (1:nu) .+ nu*(j-1)
        E[iRow, iCol] = V[iRow .- ny*(j-1),:]
        ex̂[:  , iCol] = getpower(Âpow_csum, Hp-j)*B̂u
    end
    # --- measured disturbances d ---
    gx̂ = getpower(Âpow, Hp-1)*B̂d
    G  = Matrix{Float64}(undef, Hp*ny, nd)
    jx̂ = Matrix{Float64}(undef, nx̂, Hp*nd)
    J  = repeatdiag(D̂d, Hp)
    if nd ≠ 0
        for j=1:Hp
            iRow = (1:ny) .+ ny*(j-1)
            G[iRow,:] = Ĉ*getpower(Âpow, j-1)*B̂d
        end
        for j=1:Hp
            iRow = (ny*j+1):(ny*Hp)
            iCol = (1:nd) .+ nd*(j-1)
            J[iRow, iCol] = G[iRow .- ny*j,:]
            jx̂[:  , iCol] = j < Hp ? getpower(Âpow, Hp-j-1)*B̂d : zeros(nx̂, nd)
        end
    end
    F, fx̂  = zeros(ny*Hp), zeros(nx̂) # dummy values (updated just before optimization)
    return E, F, G, J, K, V, ex̂, fx̂, gx̂, jx̂, kx̂, vx̂
end

"Return empty matrices if `model` is not a [`LinModel`](@ref)"
function init_predmat(estim::StateEstimator, model::SimModel, Hp, Hc)
    nu, nx̂, nd = model.nu, estim.nx̂, model.nd
    E  = zeros(0, nu*Hc)
    G  = zeros(0, nd)
    J  = zeros(0, nd*Hp)
    K  = zeros(0, nx̂)
    V  = zeros(0, nu)
    F  = zeros(0)
    ex̂, gx̂, jx̂, kx̂, vx̂, fx̂ = E, G, J, K, V, F
    return E, F, G, J, K, V, ex̂, fx̂, gx̂, jx̂, kx̂, vx̂
end

@doc raw"""
    init_quadprog(model::LinModel, Ẽ, S, M_Hp, N_Hc, L_Hp) -> P̃, q̃, p

Init the quadratic programming optimization matrix `P̃` and `q̃`.

The matrices appear in the quadratic general form :
```math
    J = \min_{\mathbf{ΔŨ}} \frac{1}{2}\mathbf{(ΔŨ)'P̃(ΔŨ)} + \mathbf{q̃'(ΔŨ)} + p 
```
``\mathbf{P̃}`` is constant if the model and weights are linear and time invariant (LTI). The 
vector ``\mathbf{q̃}`` and scalar ``p`` need recalculation each control period ``k`` (see
[`initpred!`](@ref) method). ``p`` does not impact the minima position. It is thus 
useless at optimization but required to evaluate the minimal ``J`` value.
"""
function init_quadprog(::LinModel, Ẽ, S, M_Hp, N_Hc, L_Hp)
    P̃ = 2*Hermitian(Ẽ'*M_Hp*Ẽ + N_Hc + S'*L_Hp*S)
    q̃ = zeros(size(P̃, 1))   # dummy value (updated just before optimization)
    p = zeros(1)            # dummy value (updated just before optimization)
    return P̃, q̃, p
end
"Return empty matrices if `model` is not a [`LinModel`](@ref)."
function init_quadprog(::SimModel, Ẽ, S, M_Hp, N_Hc, L_Hp)
    P̃ = Hermitian(zeros(0, 0))
    q̃ = zeros(0)
    p = zeros(1)            # dummy value (updated just before optimization)
    return P̃, q̃, p
end

"""
    obj_quadprog(ΔŨ, P̃, q̃)

Return the quadratic programming objective function, see [`init_quadprog`](@ref).

The function `dot(x, A, x)` is a performant way of calculating `x'*A*x`.
"""
obj_quadprog(ΔŨ, P̃, q̃) = 0.5*dot(ΔŨ, P̃, ΔŨ) + q̃'*ΔŨ

"""
    obj_nonlinprog(mpc::PredictiveController, model::LinModel, ΔŨ::Vector{Real})

Nonlinear programming objective function when `model` is a [`LinModel`](@ref).

The function is called by the nonlinear optimizer of [`NonLinMPC`](@ref) controllers. It can
also be called on any [`PredictiveController`](@ref)s to evaluate the objective function `J`
at specific input increments `ΔŨ` and predictions `Ŷ` values.
"""
function obj_nonlinprog(
    mpc::PredictiveController, model::LinModel, Ŷ, ΔŨ::Vector{T}
) where {T<:Real}
    J = obj_quadprog(ΔŨ, mpc.P̃, mpc.q̃)
    if !iszero(mpc.E)
        U = mpc.S̃*ΔŨ + mpc.T*(mpc.estim.lastu0 + model.uop)
        UE = [U; U[(end - model.nu + 1):end]]
        ŶE = [mpc.ŷ; Ŷ]
        J += mpc.E*mpc.JE(UE, ŶE, mpc.D̂E)
    end
    return J
end

"""
    obj_nonlinprog(mpc::PredictiveController, model::SimModel, ΔŨ::Vector{Real})

Nonlinear programming objective function when `model` is not a [`LinModel`](@ref). The
function `dot(x, A, x)` is a performant way of calculating `x'*A*x`.
"""
function obj_nonlinprog(
    mpc::PredictiveController, model::SimModel, Ŷ, ΔŨ::Vector{T}
) where {T<:Real}
    # --- output setpoint tracking term ---
    êy = mpc.R̂y - Ŷ
    JR̂y = dot(êy, mpc.M_Hp, êy)
    # --- move suppression and slack variable term ---
    JΔŨ = dot(ΔŨ, mpc.Ñ_Hc, ΔŨ)
    # --- input over prediction horizon ---
    if !mpc.noR̂u || !iszero(mpc.E)
        U = mpc.S̃*ΔŨ + mpc.T*(mpc.estim.lastu0 + model.uop)
    end
    # --- input setpoint tracking term ---
    if !mpc.noR̂u
        êu = mpc.R̂u - U
        JR̂u = dot(êu, mpc.L_Hp, êu)
    else
        JR̂u = 0.0
    end
    # --- economic term ---
    if !iszero(mpc.E)
        UE = [U; U[(end - model.nu + 1):end]]
        ŶE = [mpc.ŷ; Ŷ]
        E_JE = mpc.E*mpc.JE(UE, ŶE, mpc.D̂E)
    else
        E_JE = 0.0
    end
    return JR̂y + JΔŨ + JR̂u + E_JE
end

"""
    init_defaultcon(estim, C, S, N_Hc, E, ex̂, fx̂, gx̂, jx̂, kx̂, vx̂) -> con, S̃, Ñ_Hc, Ẽ

Init `ControllerConstraint` struct with default parameters based on estimator `estim`.

Also return `S̃`, `Ñ_Hc` and `Ẽ` matrices for the the augmented decision vector `ΔŨ`.
"""
function init_defaultcon(estim, Hp, Hc, C, S, N_Hc, E, ex̂, fx̂, gx̂, jx̂, kx̂, vx̂)
    model = estim.model
    nu, ny, nx̂ = model.nu, model.ny, estim.nx̂
    umin,       umax    = fill(-Inf, nu), fill(+Inf, nu)
    Δumin,      Δumax   = fill(-Inf, nu), fill(+Inf, nu)
    ymin,       ymax    = fill(-Inf, ny), fill(+Inf, ny)
    x̂min,       x̂max    = fill(-Inf, nx̂), fill(+Inf, nx̂)
    c_umin,     c_umax  = fill(0.0, nu),  fill(0.0, nu)
    c_Δumin,    c_Δumax = fill(0.0, nu),  fill(0.0, nu)
    c_ymin,     c_ymax  = fill(1.0, ny),  fill(1.0, ny)
    c_x̂min,     c_x̂max  = fill(0.0, nx̂),  fill(0.0, nx̂)
    Umin, Umax, ΔUmin, ΔUmax, Ymin, Ymax = 
        repeat_constraints(Hp, Hc, umin, umax, Δumin, Δumax, ymin, ymax)
    C_umin, C_umax, C_Δumin, C_Δumax, C_ymin, C_ymax = 
        repeat_constraints(Hp, Hc, c_umin, c_umax, c_Δumin, c_Δumax, c_ymin, c_ymax)
    A_Umin, A_Umax, S̃ = relaxU(C, C_umin, C_umax, S)
    A_ΔŨmin, A_ΔŨmax, ΔŨmin, ΔŨmax, Ñ_Hc = relaxΔU(C, C_Δumin, C_Δumax, ΔUmin, ΔUmax, N_Hc)
    A_Ymin, A_Ymax, Ẽ = relaxŶ(model, C, C_ymin, C_ymax, E)
    A_x̂min, A_x̂max, ẽx̂ = relaxterminal(model, C, c_x̂min, c_x̂max, ex̂)
    i_Umin,  i_Umax  = .!isinf.(Umin),  .!isinf.(Umax)
    i_ΔŨmin, i_ΔŨmax = .!isinf.(ΔŨmin), .!isinf.(ΔŨmax)
    i_Ymin,  i_Ymax  = .!isinf.(Ymin),  .!isinf.(Ymax)
    i_x̂min,  i_x̂max  = .!isinf.(x̂min),  .!isinf.(x̂max)
    i_b, i_g, A = init_matconstraint(
        model, 
        i_Umin, i_Umax, i_ΔŨmin, i_ΔŨmax, i_Ymin, i_Ymax, i_x̂min, i_x̂max,
        A_Umin, A_Umax, A_ΔŨmin, A_ΔŨmax, A_Ymin, A_Ymax, A_x̂max, A_x̂min
    )
    b = zeros(size(A, 1)) # dummy b vector (updated just before optimization)
    con = ControllerConstraint(
        ẽx̂      , fx̂    , gx̂     , jx̂       , kx̂     , vx̂     ,  
        Umin    , Umax  , ΔŨmin  , ΔŨmax    , Ymin   , Ymax   , x̂min   , x̂max,
        A_Umin  , A_Umax, A_ΔŨmin, A_ΔŨmax  , A_Ymin , A_Ymax , A_x̂min , A_x̂max,
        A       , b     , i_b    , C_ymin   , C_ymax , c_x̂min , c_x̂max , i_g
    )
    return con, S̃, Ñ_Hc, Ẽ
end

"Repeat predictive controller constraints over prediction `Hp` and control `Hc` horizons."
function repeat_constraints(Hp, Hc, umin, umax, Δumin, Δumax, ymin, ymax)
    Umin  = repeat(umin, Hp)
    Umax  = repeat(umax, Hp)
    ΔUmin = repeat(Δumin, Hc)
    ΔUmax = repeat(Δumax, Hc)
    Ymin  = repeat(ymin, Hp)
    Ymax  = repeat(ymax, Hp)
    return Umin, Umax, ΔUmin, ΔUmax, Ymin, Ymax
end

@doc raw"""
    relaxU(C, C_umin, C_umax, S) -> A_Umin, A_Umax, S̃

Augment manipulated inputs constraints with slack variable ϵ for softening.

Denoting the input increments augmented with the slack variable
``\mathbf{ΔŨ} = [\begin{smallmatrix} \mathbf{ΔU} \\ ϵ \end{smallmatrix}]``, it returns the
augmented conversion matrix ``\mathbf{S̃}``, similar to the one described at
[`init_ΔUtoU`](@ref). It also returns the ``\mathbf{A}`` matrices for the inequality
constraints:
```math
\begin{bmatrix} 
    \mathbf{A_{U_{min}}} \\ 
    \mathbf{A_{U_{max}}} 
\end{bmatrix} \mathbf{ΔŨ} ≤
\begin{bmatrix}
    - \mathbf{U_{min}} + \mathbf{T} \mathbf{u}(k-1) \\
    + \mathbf{U_{max}} - \mathbf{T} \mathbf{u}(k-1)
\end{bmatrix}
```
"""
function relaxU(C, C_umin, C_umax, S)
    if !isinf(C) # ΔŨ = [ΔU; ϵ]
        # ϵ impacts ΔU → U conversion for constraint calculations:
        A_Umin, A_Umax = -[S  C_umin],  [S -C_umax] 
        # ϵ has no impact on ΔU → U conversion for prediction calculations:
        S̃ = [S falses(size(S, 1))]
    else # ΔŨ = ΔU (only hard constraints)
        A_Umin, A_Umax = -S,  S
        S̃ = S
    end
    return A_Umin, A_Umax, S̃
end

@doc raw"""
    relaxΔU(C, C_Δumin, C_Δumax, ΔUmin, ΔUmax, N_Hc) -> A_ΔŨmin, A_ΔŨmax, ΔŨmin, ΔŨmax, Ñ_Hc

Augment input increments constraints with slack variable ϵ for softening.

Denoting the input increments augmented with the slack variable 
``\mathbf{ΔŨ} = [\begin{smallmatrix} \mathbf{ΔU} \\ ϵ \end{smallmatrix}]``, it returns the
augmented input increment weights ``\mathbf{Ñ}_{H_c}`` (that incorporate ``C``). It also  
returns the augmented constraints ``\mathbf{ΔŨ_{min}}`` and ``\mathbf{ΔŨ_{max}}`` and the 
``\mathbf{A}`` matrices for the inequality constraints:
```math
\begin{bmatrix} 
    \mathbf{A_{ΔŨ_{min}}} \\ 
    \mathbf{A_{ΔŨ_{max}}}
\end{bmatrix} \mathbf{ΔŨ} ≤
\begin{bmatrix}
    - \mathbf{ΔŨ_{min}} \\
    + \mathbf{ΔŨ_{max}}
\end{bmatrix}
```
"""
function relaxΔU(C, C_Δumin, C_Δumax, ΔUmin, ΔUmax, N_Hc)
    if !isinf(C) # ΔŨ = [ΔU; ϵ]
        # 0 ≤ ϵ ≤ ∞  
        ΔŨmin, ΔŨmax = [ΔUmin; 0.0], [ΔUmax; Inf]
        A_ϵ = [zeros(1, length(ΔUmin)) [1]]
        A_ΔŨmin, A_ΔŨmax = -[I  C_Δumin; A_ϵ], [I -C_Δumax; A_ϵ]
        Ñ_Hc = Diagonal([diag(N_Hc); C])
    else # ΔŨ = ΔU (only hard constraints)
        ΔŨmin, ΔŨmax = ΔUmin, ΔUmax
        I_Hc = Matrix{Float64}(I, size(N_Hc))
        A_ΔŨmin, A_ΔŨmax = -I_Hc,  I_Hc
        Ñ_Hc = N_Hc
    end
    return A_ΔŨmin, A_ΔŨmax, ΔŨmin, ΔŨmax, Ñ_Hc
end

@doc raw"""
    relaxŶ(::LinModel, C, C_ymin, C_ymax, E) -> A_Ymin, A_Ymax, Ẽ

Augment linear output prediction constraints with slack variable ϵ for softening.

Denoting the input increments augmented with the slack variable 
``\mathbf{ΔŨ} = [\begin{smallmatrix} \mathbf{ΔU} \\ ϵ \end{smallmatrix}]``, it returns the 
``\mathbf{Ẽ}`` matrix that appears in the linear model prediction equation 
``\mathbf{Ŷ = Ẽ ΔŨ + F}``, and the ``\mathbf{A}`` matrices for the inequality constraints:
```math
\begin{bmatrix} 
    \mathbf{A_{Y_{min}}} \\ 
    \mathbf{A_{Y_{max}}}
\end{bmatrix} \mathbf{ΔŨ} ≤
\begin{bmatrix}
    - \mathbf{Y_{min}} + \mathbf{F} \\
    + \mathbf{Y_{max}} - \mathbf{F} 
\end{bmatrix}
```
"""
function relaxŶ(::LinModel, C, C_ymin, C_ymax, E)
    if !isinf(C) # ΔŨ = [ΔU; ϵ]
        # ϵ impacts predicted output constraint calculations:
        A_Ymin, A_Ymax = -[E  C_ymin], [E -C_ymax] 
        # ϵ has no impact on output predictions
        Ẽ = [E zeros(size(E, 1), 1)] 
    else # ΔŨ = ΔU (only hard constraints)
        Ẽ = E
        A_Ymin, A_Ymax = -E,  E
    end
    return A_Ymin, A_Ymax, Ẽ
end

"Return empty matrices if model is not a [`LinModel`](@ref)"
function relaxŶ(::SimModel, C, C_ymin, C_ymax, E)
    Ẽ = !isinf(C) ? [E zeros(0, 1)] : E
    A_Ymin, A_Ymax = -Ẽ,  Ẽ 
    return A_Ymin, A_Ymax, Ẽ
end

@doc raw"""
    relaxterminal(::LinModel, C, c_x̂min, c_x̂max, ex̂) -> A_x̂min, A_x̂max, ẽx̂

Augment terminal state constraints with slack variable ϵ for softening.

Denoting the input increments augmented with the slack variable 
``\mathbf{ΔŨ} = [\begin{smallmatrix} \mathbf{ΔU} \\ ϵ \end{smallmatrix}]``, it returns the 
``\mathbf{ẽ_{x̂}}`` matrix that appears in the terminal state equation 
``\mathbf{x̂}_{k-1}(k + H_p) = \mathbf{ẽ_x̂ ΔŨ + f_x̂}``, and the ``\mathbf{A}`` matrices for 
the inequality constraints:
```math
\begin{bmatrix} 
    \mathbf{A_{x̂_{min}}} \\ 
    \mathbf{A_{x̂_{max}}}
\end{bmatrix} \mathbf{ΔŨ} ≤
\begin{bmatrix}
    - \mathbf{x̂_{min}} + \mathbf{f_x̂} \\
    + \mathbf{x̂_{max}} - \mathbf{f_x̂} 
\end{bmatrix}
```
"""
function relaxterminal(::LinModel, C, c_x̂min, c_x̂max, ex̂)
    if !isinf(C) # ΔŨ = [ΔU; ϵ]
        # ϵ impacts terminal constraint calculations:
        A_x̂min, A_x̂max = -[ex̂ c_x̂min], [ex̂ -c_x̂max]
        # ϵ has no impact on terminal state predictions
        ẽx̂ = [ex̂ zeros(size(ex̂, 1), 1)] 
    else # ΔŨ = ΔU (only hard constraints)
        ẽx̂ = ex̂
        A_x̂min, A_x̂max = -ex̂,  ex̂
    end
    return A_x̂min, A_x̂max, ẽx̂
end

"Return empty matrices if model is not a [`LinModel`](@ref)"
function relaxterminal(::SimModel, C, c_x̂min, c_x̂max, ex̂) 
    ẽx̂ = !isinf(C) ? [ex̂ zeros(0, 1)] : ex̂
    A_x̂min, A_x̂max = -ẽx̂,  ẽx̂
    return A_x̂min, A_x̂max, ẽx̂
end


@doc raw"""
    init_stochpred(estim::InternalModel, Hp) -> Ks, Ps

Init the stochastic prediction matrices for [`InternalModel`](@ref).

`Ks` and `Ps` matrices are defined as:
```math
    \mathbf{Ŷ_s} = \mathbf{K_s x̂_s}(k) + \mathbf{P_s ŷ_s}(k)
```
Current stochastic outputs ``\mathbf{ŷ_s}(k)`` comprises the measured outputs 
``\mathbf{ŷ_s^m}(k) = \mathbf{y^m}(k) - \mathbf{ŷ_d^m}(k)`` and unmeasured 
``\mathbf{ŷ_s^u}(k) = \mathbf{0}``. See [^2].

[^2]: Desbiens, A., D. Hodouin & É. Plamondon. 2000, "Global predictive control : a unified
    control structure for decoupling setpoint tracking, feedforward compensation and 
    disturbance rejection dynamics", *IEE Proceedings - Control Theory and Applications*, 
    vol. 147, no 4, https://doi.org/10.1049/ip-cta:20000443, p. 465–475, ISSN 1350-2379.
"""
function init_stochpred(estim::InternalModel, Hp) 
    As, B̂s, Cs = estim.As, estim.B̂s, estim.Cs
    ny  = estim.model.ny
    nxs = estim.nxs
    Ks = Matrix{Float64}(undef, ny*Hp, nxs)
    Ps = Matrix{Float64}(undef, ny*Hp, ny)
    for i = 1:Hp
        iRow = (1:ny) .+ ny*(i-1)
        Ms = Cs*As^(i-1)*B̂s
        Ks[iRow,:] = Cs*As^i - Ms*Cs
        Ps[iRow,:] = Ms
    end
    return Ks, Ps 
end
"Return empty matrices if `estim` is not a [`InternalModel`](@ref)."
init_stochpred(estim::StateEstimator, _ ) = zeros(0, estim.nxs), zeros(0, estim.model.ny)

@doc raw"""
    init_matconstraint(model::LinModel,
        i_Umin, i_Umax, i_ΔŨmin, i_ΔŨmax, i_Ymin, i_Ymax, i_x̂min, i_x̂max, args...
    ) -> i_b, i_g, A

Init `i_b`, `i_g` and `A` matrices for the linear and nonlinear inequality constraints.

The linear and nonlinear inequality constraints are respectively defined as:
```math
\begin{aligned} 
    \mathbf{A ΔŨ } &≤ \mathbf{b} \\ 
    \mathbf{g(ΔŨ)} &≤ \mathbf{0}
\end{aligned}
```
`i_b` is a `BitVector` including the indices of ``\mathbf{b}`` that are finite numbers. 
`i_g` is a similar vector but for the indices of ``\mathbf{g}`` (empty if `model` is a 
[`LinModel`](@ref)). The method also returns the ``\mathbf{A}`` matrix if `args` is
provided. In such a case, `args`  needs to contain all the inequality constraint matrices: 
`A_Umin, A_Umax, A_ΔŨmin, A_ΔŨmax, A_Ymin, A_Ymax, A_x̂min, A_x̂max`.
"""
function init_matconstraint(::LinModel, 
    i_Umin, i_Umax, i_ΔŨmin, i_ΔŨmax, i_Ymin, i_Ymax, i_x̂min, i_x̂max, args...
)
    i_b = [i_Umin; i_Umax; i_ΔŨmin; i_ΔŨmax; i_Ymin; i_Ymax; i_x̂min; i_x̂max]
    i_g = BitVector()
    if isempty(args)
        A = zeros(length(i_b), 0)
    else
        A_Umin, A_Umax, A_ΔŨmin, A_ΔŨmax, A_Ymin, A_Ymax, A_x̂min, A_x̂max = args
        A = [A_Umin; A_Umax; A_ΔŨmin; A_ΔŨmax; A_Ymin; A_Ymax; A_x̂min; A_x̂max]
    end
    return i_b, i_g, A
end

"Init `i_b, A` without outputs and terminal constraints if `model` is not a [`LinModel`](@ref)."
function init_matconstraint(::SimModel,
    i_Umin, i_Umax, i_ΔŨmin, i_ΔŨmax, i_Ymin, i_Ymax, i_x̂min, i_x̂max, args...
)
    i_b = [i_Umin; i_Umax; i_ΔŨmin; i_ΔŨmax]
    i_g = [i_Ymin; i_Ymax; i_x̂min; i_x̂max]
    if isempty(args)
        A = zeros(length(i_b), 0)
    else
        A_Umin, A_Umax, A_ΔŨmin, A_ΔŨmax, _ , _ , _ , _ = args
        A = [A_Umin; A_Umax; A_ΔŨmin; A_ΔŨmax]
    end
    return i_b, i_g, A
end

"Validate predictive controller weight and horizon specified values."
function validate_weights(model, Hp, Hc, Mwt, Nwt, Lwt, Cwt, Ewt=nothing)
    nu, ny = model.nu, model.ny
    Hp < 1  && throw(ArgumentError("Prediction horizon Hp should be ≥ 1"))
    Hc < 1  && throw(ArgumentError("Control horizon Hc should be ≥ 1"))
    Hc > Hp && throw(ArgumentError("Control horizon Hc should be ≤ prediction horizon Hp"))
    size(Mwt) ≠ (ny,) && throw(ArgumentError("Mwt size $(size(Mwt)) ≠ output size ($ny,)"))
    size(Nwt) ≠ (nu,) && throw(ArgumentError("Nwt size $(size(Nwt)) ≠ manipulated input size ($nu,)"))
    size(Lwt) ≠ (nu,) && throw(ArgumentError("Lwt size $(size(Lwt)) ≠ manipulated input size ($nu,)"))
    size(Cwt) ≠ ()    && throw(ArgumentError("Cwt should be a real scalar"))
    any(Mwt.<0) && throw(ArgumentError("Mwt weights should be ≥ 0"))
    any(Nwt.<0) && throw(ArgumentError("Nwt weights should be ≥ 0"))
    any(Lwt.<0) && throw(ArgumentError("Lwt weights should be ≥ 0"))
    Cwt < 0     && throw(ArgumentError("Cwt weight should be ≥ 0"))
    !isnothing(Ewt) && size(Ewt) ≠ () && throw(ArgumentError("Ewt should be a real scalar"))
end

function Base.show(io::IO, mpc::PredictiveController)
    Hp, Hc = mpc.Hp, mpc.Hc
    nu, nd = mpc.estim.model.nu, mpc.estim.model.nd
    nx̂, nym, nyu = mpc.estim.nx̂, mpc.estim.nym, mpc.estim.nyu
    n = maximum(ndigits.((Hp, Hc, nu, nx̂, nym, nyu, nd))) + 1
    println(io, "$(typeof(mpc).name.name) controller with a sample time Ts = "*
                "$(mpc.estim.model.Ts) s, $(solver_name(mpc.optim)) optimizer, "*
                "$(typeof(mpc.estim).name.name) estimator and:")
    println(io, "$(lpad(Hp, n)) prediction steps Hp")
    println(io, "$(lpad(Hc, n)) control steps Hc")
    print_estim_dim(io, mpc.estim, n)
end

"Limit the solving time to `mpc.model.Ts` if supported by `mpc.optim` optimizer."
function limit_solve_time(mpc::PredictiveController)
    try
        set_time_limit_sec(mpc.optim, mpc.estim.model.Ts)
    catch err
        if isa(err, MOI.UnsupportedAttribute{MOI.TimeLimitSec})
            @warn "Solving time limit is not supported by the optimizer."
        else
            rethrow(err)
        end
    end
end

"Verify that the solver termination status means 'no solution available'."
isfatal(status::TerminationStatusCode) = any(status .== FATAL_STATUSES)

"Functor allowing callable `PredictiveController` object as an alias for `moveinput!`."
function (mpc::PredictiveController)(
    ry::Vector = mpc.estim.model.yop, 
    d ::Vector = empty(mpc.estim.x̂);
    kwargs...
)
    return moveinput!(mpc, ry, d; kwargs...)
end

include("controller/explicitmpc.jl")
include("controller/linmpc.jl")
include("controller/nonlinmpc.jl")
